import sys
import os
from datetime import datetime

from tqdm.notebook import tqdm

import numpy as np
from numpy import inf
import pandas as pd

# pd.set_option('display.float_format', lambda x: '%.4f' % x)  # 在dataframe中不以科学计数法显示数值

import matplotlib.pyplot as plt
import seaborn as sns
import matplotlib as mpl

from sklearn.tree import DecisionTreeClassifier
from sklearn.metrics import roc_curve, auc

mpl.rcParams['font.family'] = 'SimHei'  # 设置中文字体
mpl.rcParams['axes.unicode_minus'] = False  # 用来正常显示负号

from pandas import read_csv, read_excel
# from westat.LogisticScoreCard import *

from westat.logger import logger
from westat.get_data_describe import get_data_describe
from westat.get_data_distribution import get_data_distribution
from westat.get_data_partition import get_data_partition
from westat.get_data_discrete import get_data_discrete
from westat.get_data_iv import get_data_iv
from westat.get_data_woe import get_data_woe
from westat.get_data_woe_transform import get_data_woe_transform
from westat.get_tree_bins import get_tree_bins
from westat.get_woe_iv import get_woe_iv, view_woe_iv
from westat.plot_woe import plot_woe
from westat.plot_iv import plot_iv

from westat.check_data_target import check_data_target
# from westat.dataframe_to_table import dataframe_to_table
from westat.get_col_type import get_col_type
from westat.get_col_bins import get_col_bins
from westat.get_modify_bins import get_modify_bins
from westat.feature_selection import get_feature_by_ivcorr

from westat.get_model_iv import get_model_iv
from westat.get_scorecard import get_scorecard

# 决策树文件转换
from westat.tree_to_img import tree_to_img
from westat.tree_to_pdf import tree_to_pdf

# 日期处理函数
from westat.date_diff import date_diff

# 设置函数别名
get_data_desc = get_data_describe
get_data_dist = get_data_distribution
get_data_part = get_data_partition


# 按切分点划分数据，得到全部的离散数据
def get_cut_result(data,
                   bins,
                   missing=[np.nan, 'nan', '', -99999999, '-99999999', -99999999.0, -99999999.0]):
    logger.info('根据cut离散化连续变量进行中。。。')
    cols = [i for i in data.columns if i not in [i[0] for i in bins]]
    data_cut_result = data[cols].copy()
    for i in tqdm(range(len(bins))):
        col = bins.iloc[i, 0]
        cut_points = bins.iloc[i, 1]
        data_cut_result[col] = pd.cut(data[col], cut_points).astype('str')

    # 缺失值替换
    data_cut_result.replace(missing, ['missing'] * len(missing), inplace=True)
    data_cut_result.fillna('missing', inplace=True)
    logging.info('根据cut离散化连续变量完成！')
    return data_cut_result


# 按变量，离散化数据
def get_col_cut_result(data, col, bins):
    col_cut_result = pd.cut(data[col], bins).astype('str')

    # 缺失值替换
    col_cut_result.replace(['nan', '', -99999999, '-99999999', -99999999.0],
                           ['missing', 'missing', 'missing', 'missing', 'missing'], inplace=True)
    col_cut_result.fillna('missing', inplace=True)
    return col_cut_result


def predict_score_proba(data, score_card, clf=False, X=False, delta_score=20, init_score=600):
    logging.info('预测用户分数中。。。')
    b = -delta_score / np.log(2)
    a = init_score - score_card['Intercept'][0] * b
    col_result = score_card['Name'].unique().tolist() + ['y']
    col_continuous_cut_points = score_card[['Name', 'Bins']][score_card['Type'] == 'continuous'].drop_duplicates('Name')
    data_discrete = get_cut_result(data[col_result], col_continuous_cut_points)
    data_score_proba = pd.DataFrame()
    for col in score_card['Name'].unique():
        col_score = col + '_score'
        cut_points = score_card['Bin'][score_card['Name'] == col].tolist()
        score = score_card['Score'][score_card['Name'] == col].tolist()
        data_score_proba[col_score] = data_discrete[col].replace(cut_points, score)
    data_score_proba['y'] = data_discrete['y']
    data_score_proba['Score'] = data_score_proba.sum(axis=1) + score_card['Intercept'][0] * b + a
    if clf:
        data_score_proba['Proba'] = clf.predict_proba(X)[:, 1]
    else:
        data_score_proba['Proba'] = 1 - 1 / (1 + np.e ** ((data_score_proba['Score'] - a) / b))
    return data_score_proba


def plot_roc_ks(data, score_card):
    data_score_proba = predict_score_proba(data, score_card)
    false_positive_rate, recall, thresholds = roc_curve(data['y'], data_score_proba['Proba'], drop_intermediate=False)
    roc_auc = auc(false_positive_rate, recall)
    plt.figure(figsize=(20, 10))

    # ROC曲线
    plt.subplot(121)
    plt.title('ROC')
    plt.plot(false_positive_rate, recall, 'b', label='AUC = %0.4f' % roc_auc)
    plt.legend(loc='lower right')
    plt.plot([0, 1], [0, 1], 'r--')
    plt.xlim([0.0, 1.0])
    plt.ylim([0.0, 1.0])
    plt.xlabel('false_positive_rate')
    plt.ylabel('recall')

    # KS曲线
    plt.subplot(122)
    pre = sorted(data_score_proba['Proba'], reverse=True)
    num = [(i) * int(len(pre) / 10) for i in range(10)]
    num = num + [(len(pre) - 1)]
    ks_thresholds = [max(thresholds[thresholds <= pre[i]]) for i in num]
    data_ks = pd.DataFrame([false_positive_rate, recall, thresholds]).T
    data_ks.columns = ['fpr', 'tpr', 'thresholds']
    data_ks = pd.merge(data_ks, pd.DataFrame(ks_thresholds, columns=['thresholds']), on='thresholds', how='inner')
    ks = max(recall - false_positive_rate)
    plt.title('KS')
    plt.plot(np.array(range(len(num))), data_ks['tpr'])
    plt.plot(np.array(range(len(num))), data_ks['fpr'])
    plt.plot(np.array(range(len(num))), data_ks['tpr'] - data_ks['fpr'], label='K-S = %0.4f' % ks)
    plt.legend(loc='lower right')
    plt.xlim([0, 10])
    plt.ylim([0.0, 1.0])
    plt.xlabel('label')
    plt.show()


def get_score_dist(data, score_card, qcut=10, precision=2, language='en'):
    result_score = predict_score_proba(data, score_card)

    if qcut:
        result_score['Score'] = pd.qcut(result_score['Score'], qcut)

    result = result_score.groupby('Score')['y'].agg([('#Bad', lambda y: (y == 1).sum()),
                                                     ('#Good', lambda y: (y == 0).sum()),
                                                     ('#Total', 'count')]).reset_index()
    result['%Bad'] = result['#Bad'] / result['#Bad'].sum()
    result['%Good'] = result['#Good'] / result['#Good'].sum()
    result['%Total'] = result['#Total'] / result['#Total'].sum()
    result['%BadRate'] = result['#Bad'] / result['#Total']

    result['#CumBad'] = result['#Bad'].cumsum()
    result['#CumGood'] = result['#Good'].cumsum()
    result['#CumTotal'] = result['#Total'].cumsum()
    result['%CumBad'] = result['#CumBad'] / result['#Bad'].sum()
    result['%CumGood'] = result['#CumGood'] / result['#Good'].sum()
    result['LIFT'] = (result['#Bad'] / result['#Total']) / (result['#Bad'].sum() / result['#Total'].sum())

    precision_str = '.' + str(precision)
    result['%Total'] = result['%Total'].apply(lambda x: format(x, precision_str + '%'))
    result['%Bad'] = result['%Bad'].apply(lambda x: format(x, precision_str + '%'))
    result['%Good'] = result['%Good'].apply(lambda x: format(x, precision_str + '%'))
    result['%BadRate'] = result['%BadRate'].apply(lambda x: format(x, precision_str + '%'))
    result['%CumBad'] = result['%CumBad'].apply(lambda x: format(x, precision_str + '%'))
    result['%CumGood'] = result['%CumGood'].apply(lambda x: format(x, precision_str + '%'))
    result['LIFT'] = result['LIFT'].apply(lambda x: round(x, precision))
    result['NO.'] = result.index + 1

    result = result[
        ['NO.', 'Score', '#Total', '#Bad', '#Good', '%Total', '%Bad', '%Good', '%BadRate', '%CumBad', 'LIFT']]

    return result


def view_score_dist(data, score_card, qcut=10, precision=2):
    result = get_score_dist(data=data, score_card=score_card, qcut=qcut, precision=precision)
    result['LIFT.'] = result['LIFT']
    result = result.style.bar(subset=['LIFT.'], color='#007bff')
    return result


class Table(pd.DataFrame):
    def __init__(self):
        super().__init__()

    def __str__(self):
        return 'table for westat'



import numpy as np
import pandas as pd
from tqdm.notebook import tqdm
from westat.logger import logger
from westat.get_woe_iv import get_woe_iv
from westat.get_data_iv import get_data_iv

def get_feature_by_ivcorr(data: pd.DataFrame,    
                          col_iv: pd.DataFrame = pd.DataFrame(),                   
                          min_iv: float = 0.02,
                          max_corr: float = 0.6,
                          keep: list = [],
                          drop: list = [],
                          target: str = 'target',
                          return_drop: bool = False):
    """
    根据最小IV值 和 最大相关性 筛选特征
    Args:
        data:DataFrame,将要筛选特征的数据集
        col_iv:DataFrame,列的IV汇总，包含Name,IV 两列的数据集
        min_iv:筛选后允许的最小IV值
        max_corr:筛选后允许的最大相关性
        keep:list,需要保留的特征
        drop:list,需要删除的特征
        target:str,目标变量名称，默认为'y'
        return_drop:是否返回已删除的特征

    Returns:
        默认返回筛选后的特征名单，
        当return_drop为True时，同时返回筛选后的特征名单、删除的特征名单、相关性和IV数据表
    """
    if col_iv.empty:
        col_iv = get_data_iv(data, target=target)

    # 根据每一列的IV值，计算相关矩阵
    col_iv_filter = col_iv[col_iv['IV'] >= min_iv]
    data_corr = data[col_iv_filter['Name']].corr()

    col_iv_result = []
    for i in range(len(col_iv_filter)):
        for j in range(len(col_iv_filter)):
            col1 = col_iv_filter.iloc[i, 0]
            iv1 = col_iv_filter.iloc[i, 1]
            col2 = col_iv_filter.iloc[j, 0]
            iv2 = col_iv_filter.iloc[j, 1]
            col_iv_result.append([col1, col2, iv1, iv2, iv1 - iv2])

    result = pd.DataFrame(col_iv_result, columns=['Name1', 'Name2', 'IV1', 'IV2', 'IV1-IV2'])
    result['Corr'] = data_corr.values.reshape(-1, 1)
    result = result[result['Name1'] != result['Name2']]

    col_delete_iv_corr = result['Name1'][(result['IV1'] <= min_iv) | (
            (result['Corr'] > max_corr) & (result['IV1-IV2'] < 0))].unique()

    col_result = [col for col in result['Name1'].unique() if col not in col_delete_iv_corr and col not in drop]
    col_keep = sorted(list(set(col_result + keep)))
    col_drop = sorted(list(set(list(col_delete_iv_corr) + drop)))

    # 是否返回已删除特征
    if return_drop:
        return col_keep, col_drop, result
    else:
        return col_keep



from westat import pd
from westat import tqdm
from westat.get_woe_iv import get_woe_iv


def get_model_iv(data_discrete,
                 data_iv,
                 col_bins,
                 col_dict,
                 col_types,
                 target='y',
                 precision=2,
                 language='en'):
    result = pd.DataFrame()
    for col in tqdm(data_discrete.columns):
        if col != target and data_iv['IV'][data_iv['Name'] == col].values[0] != 0:
            df_woe = get_woe_iv(data=data_discrete, col=col, criterion='discrete', target=target, precision=precision)
            result = pd.concat([result, df_woe])

    # 如果数据字典不为空，则填充字段说明，否则字段说明为空
    if col_dict.empty:
        result_1 = result
        result_1['Label'] = ''
    else:
        result_1 = result.merge(col_dict, on='Name', how='left')
        result_1['Label'] = result_1.iloc[:, -1]

        # 合并数据类型
    result_2 = result_1.merge(col_types, on='Name', how='left')

    # 合并数据分箱
    result = result_2.merge(col_bins, on='Name', how='left')
    result.sort_values(by=['Total IV', 'Name', 'No.'], ascending=[False, True, True], inplace=True)
    result.reset_index(drop=True, inplace=True)

    result['WoE.'] = result['WoE']
    result['Style'] = result['Name'].apply(lambda x: list(result['Name'].unique()).index(x) % 2)

    # 语言设置
    if language == 'cn':
        result = result[['Name', 'Label', 'Type', 'No.', 'Bin', 'Bins', '#Total', '#Bad', '#Good', '%Total', '%Bad',
                         '%Good', '%BadRate', 'WoE', 'IV', 'Total IV', 'WoE.', 'Style']]
        result.rename(
            columns={'Name': '名称', 'Label': '描述', 'Type': '类型', 'No.': '分箱序号', 'Bin': '分箱', 'Bins': '切分点',
                     '#Total': '#合计', '#Bad': '#坏', '#Good': '好', '%Total': '%合计', '%Bad': '%坏', '%Good': '%好',
                     '%BadRate': '%坏件率', 'Total IV': 'IV合计', 'Style': '样式'}, inplace=True)
    else:
        result = result[['Name', 'Label', 'Type', 'No.', 'Bin', 'Bins', '#Total', '#Bad', '#Good', '%Total', '%Bad',
                         '%Good', '%BadRate', 'WoE', 'IV', 'Total IV', 'WoE.', 'Style']]
    return result
